#include <QtGui>

#include "pfdisk.h"
#include <cstdio>
#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/stat.h>
#include <cstring>
#include <unistd.h>
#include <cstdlib>
#include <cctype>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <libdevinfo.h>
#include <dirent.h>
#include <strings.h> // bzero
#include <sys/ioctl.h>
#include <sys/dktp/fdisk.h>
#include <iostream>
#include <map>
using std::strcat;
using std::strcpy;
using std::sprintf;
using std::strcmp;
using std::exit;
using std::cerr;
using std::endl;
using std::vector;
using std::map;
using std::make_pair;

std::map<int, QString> gIdNameMap;
#define MAP_ADD(x) gIdNameMap.insert(make_pair(x, QString(#x)))

pfdisk::pfdisk(QWidget */*parent*/)
{
    setupUi(this);
    MAP_ADD(DOSOS12);
    MAP_ADD(PCIXOS);
    MAP_ADD(DOSOS16);
    MAP_ADD(EXTDOS);
    MAP_ADD(DOSHUGE);
    MAP_ADD(FDISK_IFS);
    MAP_ADD(FDISK_AIXBOOT);
    MAP_ADD(FDISK_AIXDATA);
    MAP_ADD(FDISK_OS2BOOT);
    MAP_ADD(FDISK_WINDOWS);
    MAP_ADD(FDISK_EXT_WIN);
    MAP_ADD(FDISK_FAT95);
    MAP_ADD(FDISK_EXTLBA);
    MAP_ADD(DIAGPART);
    MAP_ADD(FDISK_LINUX);
    MAP_ADD(FDISK_LINUXDSWAP);
    MAP_ADD(FDISK_LINUXDNAT);
    MAP_ADD(FDISK_CPM);
    MAP_ADD(DOSDATA);
    MAP_ADD(OTHEROS);
    MAP_ADD(UNIXOS);
    MAP_ADD(UNUSED);
    MAP_ADD(FDISK_NOVELL3);
    MAP_ADD(FDISK_QNX4);
    MAP_ADD(FDISK_QNX42);
    MAP_ADD(FDISK_QNX43);
    MAP_ADD(SUNIXOS);
    MAP_ADD(FDISK_LINUXNAT);
    MAP_ADD(FDISK_NTFSVOL1);
    MAP_ADD(FDISK_NTFSVOL2);
    MAP_ADD(FDISK_BSD);
    MAP_ADD(FDISK_NEXTSTEP);
    MAP_ADD(FDISK_BSDIFS);
    MAP_ADD(FDISK_BSDISWAP);
    MAP_ADD(X86BOOT);
    MAP_ADD(SUNIXOS2);
    MAP_ADD(EFI_PMBR);
    MAP_ADD(EFI_FS);

    mValidator6 = new ODHValidator(0, 1<<5, this, "6 bit validator");
    mValidator8 = new ODHValidator(0, 1<<7, this, "8 bit validator");
    mValidator8m1 = new ODHValidator(0, 1<<7-1, this, "8 bit validator, minus 1");
    mValidator10 = new ODHValidator(0, 1<<9, this, "10 bit validator");
    mValidator32 = new ODHValidator(0, 1<<31, this, "32 bit validator");

    for (int i = 0; i < 4; ++i)
    {
        mStateLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1StateLineEdit").arg(i+1)));
        
        mCylinderStartLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1CylinderStartLineEdit").arg(i+1)));
        mHeadStartLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1HeadStartLineEdit").arg(i+1)));
        mSectorStartLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1SectorStartLineEdit").arg(i+1)));

        mTypeLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1TypeLineEdit").arg(i+1)));
        //        new NameMapTip(mTypeLineEdit[i]);

        mCylinderEndLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1CylinderEndLineEdit").arg(i+1)));
        mHeadEndLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1HeadEndLineEdit").arg(i+1)));
        mSectorEndLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1SectorEndLineEdit").arg(i+1)));

        mLBAStartLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1LBAStartLineEdit").arg(i+1)));
        mLBALengthLineEdit[i] = dynamic_cast<QLineEdit*>(child(QString("entry%1LBALengthLineEdit").arg(i+1)));
    }
    
    mInitDone = false;
    startHeadTextLabel->setMinimumWidth(startCylinderTextLabel->sizeHint().width());
    startSectorTextLabel->setMinimumWidth(startCylinderTextLabel->sizeHint().width());
    endHeadTextLabel->setMinimumWidth(endCylinderTextLabel->sizeHint().width());
    endSectorTextLabel->setMinimumWidth(endCylinderTextLabel->sizeHint().width());

    // these need to be big enough to hold 11 digits
    int w = entry1LBAStartLineEdit->fontMetrics().width("377777777777");
    entry1LBAStartLineEdit->setMinimumWidth(w);
    entry1LBALengthLineEdit->setMinimumWidth(w);
    getAlias("sd");
    getAlias("cmdk");

    if (mDeviceArray.empty())
    {
        QMessageBox::warning(this, tr("Warning"), tr("No hard disks found [sd and cmdk]"));
        return;
    } // end if
    
    mPartitionTables.resize(mDeviceArray.size());
    
    int f, index = 0;
    char buf[512];
    for (vector<QString>::const_iterator iter = mDeviceArray.begin();
         iter != mDeviceArray.end();
         ++iter, ++index)
    {
	dasdComboBox->insertItem(*iter);

        f = ::open(iter->local8Bit(), O_RDONLY);
        if (!f)
        {
            cerr << "failed to open " << iter->toLocal8Bit().constData() << " " << strerror(errno) << endl;
            return;
        } // end if
        int r = ::read(f, buf, 512);

        if (r != 512)
        {
            cerr << "failed to read from " << iter->toLocal8Bit().constData() << " " << strerror(errno) << endl;
            return;
        } // end if

        for (int i = 0; i < 4; ++i)
        {
            ::memcpy(&mPartitionTables[index].mTable[i], &buf[0x1be + 16*i], 16);
            extractCylinderSector(mPartitionTables[index].mTable[i]);
        } // end for

        ::close(f);
    } // end for

    for (int i = 0; i < 4; ++i)
    {
        mStateLineEdit[i]->setValidator(mValidator8);
        mHeadStartLineEdit[i]->setValidator(mValidator8m1);
        mTypeLineEdit[i]->setValidator(mValidator8);
        mHeadEndLineEdit[i]->setValidator(mValidator8m1);
        mCylinderStartLineEdit[i]->setValidator(mValidator10);
        mCylinderEndLineEdit[i]->setValidator(mValidator10);
        mSectorStartLineEdit[i]->setValidator(mValidator6);
        mSectorEndLineEdit[i]->setValidator(mValidator6);
        mLBAStartLineEdit[i]->setValidator(mValidator32);
        mLBALengthLineEdit[i]->setValidator(mValidator32);
    } // end for
    mCurrentDisk = 0;
    changeBase(1);
    actionTypeChanged(QString("ignored"));
    mInitDone = true;
}

pfdisk::~pfdisk()
{
    delete mValidator6;
    delete mValidator8;
    delete mValidator8m1;
    delete mValidator10;
    delete mValidator32;
}

/* 
 * This routine is used to traverse the devinfo tree looking
 * for devices using the driver specified in the argument
 * list and obtaining the physical path information about them.
 */

int pfdisk::getAlias(const QString& alias)
{
    QString file;
    QString value;
    di_node_t root_node;
    di_node_t node;
    di_minor_t minor = DI_MINOR_NIL;
    QString tmpnode;
    char resolvedPath[MAXPATHLEN];
    char* phys_path;
    char* minor_name;
    int instance;

    if ((root_node = ::di_init("/", DINFOCPYALL)) == DI_NODE_NIL)
    {
        cerr << "di_init() failed " << strerror(errno) << endl;
        return -1;
    }
    node = ::di_drv_first_node(alias.local8Bit(), root_node);
    while (node != DI_NODE_NIL)
    {

        if ((minor = ::di_minor_next(node, DI_MINOR_NIL)) != DI_MINOR_NIL)
        {
            scsi_inquiry inq_data;
            int fd;
            instance = ::di_instance(node);
            phys_path = ::di_devfs_path(node);
            minor_name = ::di_minor_name(minor);

            tmpnode = alias + QString::number(instance);
            file = "/devices";
            file += phys_path;
            file += ":";
            file += minor_name;
            file += ",raw";

            if (-1 == readDirectory(file, value))
            {
                cerr << "readDirectory failed" << endl;
                continue;
            }
            //printf("Device Driver Instance Name = %8s, SVR4 name = %s\n", tmpnode, value);

            int rlength;
            if (-1 == (rlength = ::resolvepath(value.local8Bit(), resolvedPath, MAXPATHLEN)))
            {
                cerr << "resolvepath() failed " << strerror(errno) << endl;
                return -1;
            }
            else
            {
                resolvedPath[rlength] = '\0';
            } // end if

            if ("sd" == alias)
            {
                if ((fd = ::open(resolvedPath, O_RDONLY|O_NDELAY)) < 0)
                {
                    cerr << "cannot open " << resolvedPath << " " << strerror(errno) << endl;
                    fd = 0;
                }
                else
                {
                    if (inquiry(fd, inq_data, sizeof(inq_data)))
                    {
                        cerr << "scsi inquiry command failed" << endl;
                        ::close(fd);
                        return -1;
                    } // end if
                } // end if
            } // end if

            if ("sd" != alias || (fd && !inq_data.inq_rmb))
            {
                QString device(value);
                device.replace(device.length()-2, 1, QChar('p'));
                mDeviceArray.push_back(device);
            } // end if
            
            ::di_devfs_path_free(phys_path);
            node = ::di_drv_next_node(node);
            ::bzero(resolvedPath, MAXPATHLEN);

            if(fd > 0)
            {
                ::close(fd);
                fd = 0;
            } // end if
        }
        else
        {
            node = ::di_drv_next_node(node);
        } // end if
    } /* End of the while loop */

    ::di_fini(root_node);
    return 0;
}

/* This routine just reads the link of each entry in the
 * /dev/rdsk directory and compares it with the physical
 * path information recieved from the caller routine. If
 * they match and the file is a block device, the routine
 * updates the value character array and returns 0. Else
 * it returns a -1.
 */
int pfdisk::readDirectory(const QString& diskPath, QString& value)
{
    DIR     *dirp;
    struct  dirent *dp;
    struct  stat stbuf;
    QString currentDir("/dev/rdsk");
    char fileName[MAXPATHLEN];
    QString tempName;
    QString dirDName;

    if ((dirp = ::opendir(currentDir.local8Bit())) != 0)
    {
        while ((dp = ::readdir(dirp)) != 0)
        {
            // this is made from the name extracted with the devinfo calls
            tempName = "../..";
            tempName += diskPath;
            // this is the name in the directory
            dirDName = "/dev/rdsk/";
            dirDName += dp->d_name;
            int statStatus = ::stat(dirDName.local8Bit(), &stbuf);
            if (-1 == statStatus)
            {
                continue;
            }

            // check to see if the file is char mode
            if (S_ISCHR(stbuf.st_mode))
            {
                // read the link
                ssize_t readlinkStatus = ::readlink(dirDName.local8Bit(), fileName, sizeof(fileName));
                if (-1 == readlinkStatus)
                {
                    cerr << "readlink failed on " << dirDName.toLocal8Bit().constData() << " " << strerror(errno) << endl;
                    continue;
                } // end if

                // compare it to the devinfo one
                if (fileName == tempName)
                {
                    // success
                    value = dirDName;
                    ::closedir(dirp);
                    return 0;
                } // end if
            } // end if

            ::bzero(fileName, MAXPATHLEN);

        } // end while
        ::closedir(dirp);
    } // end if

    return(-1);
}



int pfdisk::inquiry(int fd,  scsi_inquiry& inq_data, int inq_len)
{
    struct uscsi_cmd sc;
    union scsi_cdb inq_cdb;

    ::memset(&inq_cdb, 0, sizeof(inq_cdb));
    inq_cdb.scc_cmd = SCMD_INQUIRY;
    inq_cdb.g0_count0 = inq_len;

    ::memset(&sc, 0, sizeof(sc));
    sc.uscsi_cdb = (caddr_t)&inq_cdb;
    sc.uscsi_cdblen = 6;
    sc.uscsi_bufaddr = (char*)&inq_data;
    sc.uscsi_buflen = inq_len;
    sc.uscsi_flags = USCSI_ISOLATE|USCSI_READ;
    
    if (::ioctl(fd, USCSICMD, &sc))
    {
        cerr << "USCSICMD ioctl, inquiry " << strerror(errno) << endl;
        return -1;
    }

    if (sc.uscsi_status)
    {
        cerr << "bad status for INQUIRY command (" << sc.uscsi_status << ")" << endl;
        return -1;
    }
    return 0;
}



void pfdisk::refresh(int index)
{
    if (index != mCurrentDisk)
    {
        readCurrentDisk();
        mCurrentDisk = index;
    } // end if

    for (int i = 0; i < 4; ++i)
    {    
        mStateLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].state, mBase));
        mHeadStartLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].headStart, mBase));
        mCylinderStartLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].cylinderSectorStart.extractedData.cylinder, mBase));
        mSectorStartLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].cylinderSectorStart.extractedData.sector, mBase));
        mTypeLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].type, mBase));
        mHeadEndLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].headEnd, mBase));
        mCylinderEndLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].cylinderSectorEnd.extractedData.cylinder, mBase));
        mSectorEndLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].cylinderSectorEnd.extractedData.sector, mBase));
        mLBAStartLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].lbaStart, mBase));
        mLBALengthLineEdit[i]->setText(QString::number(mPartitionTables[index].mTable[i].lbaLength, mBase));
    } // end for
}



void pfdisk::changeBase(int index)
{
    if (mInitDone)
    {
        readCurrentDisk();
    } // end if

    switch (index)
    {
    case 0:
        // state 3 digits
        // head start/end 3 digits
        // cylinder start/end 4 digits
        // sector start/end 2 digits
        // type 3 digits
        // LBA start/length 11 digits
        mBase = 8;
        break;
    case 1:
    default:
        // state 3 digits
        // head start/end 3 digits
        // cylinder start/end 4 digits
        // sector start/end 2 digits
        // type 3 digits
        // LBA start/length 10 digits
        mBase = 10;
        break;
    case 2:
        // state 2 digits
        // head start/end 2 digits
        // cylinder start/end 3 digits
        // sector start/end 2 digits
        // type 2 digits
        // LBA start/length 8 digits
        mBase = 16;
        break;
    } // end switch
    
    mValidator6->setMode((ODHValidator::tMode)index);
    mValidator8->setMode((ODHValidator::tMode)index);
    mValidator8m1->setMode((ODHValidator::tMode)index);
    mValidator10->setMode((ODHValidator::tMode)index);
    mValidator32->setMode((ODHValidator::tMode)index);

    refresh(dasdComboBox->currentItem());
}



void pfdisk::readCurrentDisk()
{
    //if (!mInitDone)
    //{
    //    return;
    //} // end if
    bool ok;
    for (int i = 0; i < 4; ++i)
    {
        mPartitionTables[mCurrentDisk].mTable[i].state = mStateLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].headStart = mHeadStartLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].cylinderSectorStart.extractedData.cylinder = mCylinderStartLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].cylinderSectorStart.extractedData.sector = mSectorStartLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].type = mTypeLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].headEnd = mHeadEndLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].cylinderSectorEnd.extractedData.cylinder = mCylinderEndLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].cylinderSectorEnd.extractedData.sector = mSectorEndLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].lbaStart = mLBAStartLineEdit[i]->text().toUInt(&ok, mBase);
        mPartitionTables[mCurrentDisk].mTable[i].lbaLength = mLBALengthLineEdit[i]->text().toUInt(&ok, mBase);
    } // end for
}


void pfdisk::actionAbout()
{
    QMessageBox* m = new QMessageBox(this, "about message box");
    m->setText(tr("Solaris x86 Partition Editor Copyright &copy;<br>Paul John Floyd<br>2004 - 2007"));
    m->setIcon(QMessageBox::Information);
    // what option???
    //m->setTextFormat(RichText);
    m->setCaption(tr("About pfdisk"));
    m->show();
}


void pfdisk::actionSave()
{
    // ask to make backup
    // are you really sure?
    
    switch (QMessageBox::question(this, tr("Make Backup"),
                                  tr("Do you want to make a backup of this partition table?"),
                                  tr("Yes"), tr("No"), tr("Cancel")))
    {
    case 0:
        makeBackup();
        break;
    case 1:
        break;
    case 2:
    default:
        return;
    } // end switch
    
    switch (QMessageBox::critical(this, tr("Final Warnning!"),
                                  tr("Do you want to continue?"),
                                  tr("Yes"), tr("No")))
    {
    case 0:
        break;
    case 1:
    default:
        return;
    } // end switch

    // OK, do it
    readCurrentDisk();
    PartitionTable outTable = mPartitionTables[dasdComboBox->currentItem()];
    // convert back to PT format
    for (int i = 0; i < 4; ++i)
    {
        restoreCylinderSector(outTable.mTable[i]);
    } // end for

    // open partiton
    char buf[512];
    int f = ::open(dasdComboBox->currentText().local8Bit(), O_RDWR);
    if (!f)
    {
        cerr << "failed to open " << dasdComboBox->currentText().toLocal8Bit().constData() << " " << strerror(errno) << endl;
        return;
    } // end if

    // read in
    int r = ::read(f, buf, 512);

    if (r != 512)
    {
        cerr << "failed to read from " << dasdComboBox->currentText().toLocal8Bit().constData() << " " << strerror(errno) << endl;
        ::close(f);
        return;
    } // end if

    // copy partiton table into buffer
    for (int i = 0; i < 4; ++i)
    {
        ::memcpy(&buf[0x1be + 16*i], &outTable.mTable[i], 16);
    } // end for

    // rewind partition
    if (-1 == ::lseek(f, 0, SEEK_SET))
    {
        cerr << "failed to seek to start of partition " << strerror(errno) << endl;
        return;
    } // end if
            

    // write out the buffer
    int w = ::write(f, buf, 512);

    if (w != 512)
    {
        cerr << "failed to write buffer to " << dasdComboBox->currentText().toLocal8Bit().constData() << " " << strerror(errno) << endl;
        ::close(f);
        return;
    } // end if    

    ::close(f);
}

void pfdisk::actionTypeChanged(QString /*newString*/)
{
    for (int i = 0; i < 4; ++i)
    {
        mTypeLineEdit[i]->setToolTip(gIdNameMap[mTypeLineEdit[i]->text().toInt()]);
    }
}

void pfdisk::makeBackup()
{
    QString backupFile = QFileDialog::getSaveFileName(QFileInfo(dasdComboBox->currentText()).fileName(),
                                                      tr("*.pt"),
                                                      this,
                                                      "save_pt_getsavefilename");

    PartitionTable pt = mPartitionTables[dasdComboBox->currentItem()];
    for (int i = 0; i < 4; ++i)
    {
        restoreCylinderSector(pt.mTable[i]);
    } // end for

    int f = ::open(backupFile.local8Bit(), O_WRONLY|O_CREAT);
    if (!f)
    {
        std::cerr << "failed to open " << backupFile.toLocal8Bit().constData() << " " << strerror(errno) << std::endl;
        return;
    } // end if

    if (sizeof(pt) != ::write(f, &pt, sizeof(pt)))
    {
        cerr << "failed to write all data to " << backupFile.toLocal8Bit().constData() << " " << strerror(errno) << endl;
        ::close(f);
        return;
    } // end if

    ::close(f);
}



void pfdisk::actionDiagnosis()
{
    readCurrentDisk();
    DiagnosisWizard* dw = new DiagnosisWizard(this);
    dw->setData(&mPartitionTables, &mDeviceArray);
    dw->exec();
    refresh(mCurrentDisk);
}
