// Copyright Paul Floyd 2005 - 2006

#include "prime_factor.h"

#include <iostream>
#include <algorithm>
#include <numeric> // accumulate
#include <functional> // multiplies
#include <cmath>
#include <cstdlib> // atol
using std::cout;
using std::endl;
using std::vector;
using std::sqrt;
using std::atol;
using std::find;
using std::multiplies;
using std::accumulate;
using std::for_each;

std::vector<unsigned long int> factor(unsigned long int value)
{
    unsigned long int root = static_cast<unsigned long int>(sqrt(static_cast<double>(value)));
    unsigned long int firsttwo[] = { 2ul, 3ul };

    static vector<unsigned long int> primes = vector<unsigned long int>(&firsttwo[0],
                                                                        &firsttwo[2]);
    vector<unsigned long int> factors;

    unsigned long int i = primes.back();
    bool isPrime;
    while (root >= primes.back())
    {
        i += 2;
        isPrime = true;
        for (vector<unsigned long int>::const_iterator iter = primes.begin();
             iter != primes.end();
             ++iter)
        {
            if (!(i % *iter))
            {
                isPrime = false;
            } // end if
        } // end for
        if (isPrime)
        {
            primes.push_back(i);
        }
    } // end while

    // Knuth Algorithm A, p.380 Seminumerical Algorithms
    // A1
    unsigned long int t = 0;
    unsigned long int k = 0;
    unsigned long int n = value;
    unsigned long int q = 0;
    unsigned long int r = 0;

    while (true)
    {
        // A2
        if (n == 1)
        {
            break;
        }

        //A3:        
        // A3
        q = n/primes[k];
        r = n%primes[k];

        if (r)
        {
            // A6
            if (q > primes[k])
            {
                ++k;
                //goto A3;
                continue;
            }
            
        }
        else
        {
            // A5
            ++t;
            factors.push_back(primes[k]);
            n = q;
            continue;
        }

        // A7
        ++t;
        factors.push_back(n);
        break;
    } // end while

    return factors;
}

void PrintIt(unsigned long int n)
{
    cout << n  << " " ;
}

void getCHS(unsigned long int& c, unsigned long int& h, unsigned long int&s, unsigned long int address)
{
    vector<unsigned long int> address_factors = factor(address);
    vector<unsigned long int> factors;

    const int startValues[] = { 63, 255 };
    unsigned long int results[2];

    for (int x = 0; x < 2; ++x)
    {
        unsigned long int i;
        for (i = startValues[x]; i >= 2; --i)
        {
            factors = factor(i);

            // if all elements of i_factors are in address_factors,
            // then we've done
            // I don't think that there's and standard algorithm
            // to do this sort of search
            bool found = true;
            vector<unsigned long int>::iterator afci = address_factors.begin();
            for (vector<unsigned long int>::const_iterator fci = factors.begin();
                 fci != factors.end();
                 ++fci)
            {
                if (address_factors.end() == (afci = find(afci,
                                                          address_factors.end(),
                                                          *fci)))
                {
                    found = false;
                    break;
                }
                else
                {
                    ++afci;
                } // end if
            } // end for

            if (found)
            {
                for (vector<unsigned long int>::const_iterator fci = factors.begin();
                     fci != factors.end();
                     ++fci)
                {
                    address_factors.erase(find(address_factors.begin(),
                                               address_factors.end(),
                                               *fci));
                } // end for
                break;
            } // end if
        } // end for

        if (1 == i)
        {
            results[x] = 1;
        }
        else
        {
            results[x] = accumulate(factors.begin(), factors.end(), 1ul, multiplies<unsigned long int>());
        } // end if
    } // end for

    s = results[0];
    h = results[1];
    c = accumulate(address_factors.begin(), address_factors.end(), 1ul, multiplies<unsigned long int>());
    return;
}

